"""Numerical quadrature node and weight generator."""
import logging
import numpy
import chaospy

from .utils import combine
from .sparse_grid import sparse_grid

SHORT_NAME_TABLE = {
    "c": "clenshaw_curtis", "clenshaw_curtis": "clenshaw_curtis",
    "f1": "fejer_1", "fejer_1": "fejer_1",
    "f2": "fejer_2", "fejer_2": "fejer_2",
    "g": "gaussian", "gaussian": "gaussian",
    "e": "legendre", "legendre": "legendre",
    "l": "lobatto", "lobatto": "lobatto",
    "k": "kronrod", "kronrod": "kronrod",
    "p": "patterson", "patterson": "patterson",
    "r": "radau", "radau": "radau",
    "j": "leja", "leja": "leja",
    "n": "newton_cotes", "newton_cotes": "newton_cotes",
    "d": "discrete", "discrete": "discrete",
    "i": "grid", "grid": "grid",
    "z16": "genz_keister_16", "genz_keister_16": "genz_keister_16",
    "z18": "genz_keister_18", "genz_keister_18": "genz_keister_18",
    "z22": "genz_keister_22", "genz_keister_22": "genz_keister_22",
    "z24": "genz_keister_24", "genz_keister_24": "genz_keister_24",
}
DEPRECATED_SHORT_NAMES = {
    "f": "f2",
    "fejer": "fejer_2",
    "gauss_kronrod": "kronrod",
    "gauss_lobatto": "lobatto",
    "gauss_patterson": "patterson",
    "gauss_radau": "radau",
    "gauss_legendre": "legendre",
    "z": "genz_keister_24",
    "genz_keister": "genz_keister_24",
}


def generate_quadrature(
        order,
        dist,
        rule=None,
        sparse=False,
        growth=None,
        segments=1,
        recurrence_algorithm="stieltjes",
        tolerance=1e-10,
        scaling=3,
        n_max=5000,
):
    """
    Numerical quadrature node and weight generator.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution which density will be used as weight function.
        rule (str, Sequence[str], None):
            Rule for generating abscissas and weights. If one name is provided,
            that rule is applied to all dimensions. If multiple names, each
            rule is positionally applied to each dimension. If omitted,
            ``clenshaw_curtis`` is applied to all continuous dimensions, and
            ``discrete`` to all discrete ones.
        sparse (bool):
            If True used Smolyak's sparse grid instead of normal tensor product
            grid.
        growth (bool, None):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples. Defaults to the same
            value as ``sparse`` if omitted.
        segments (int):
            Split intervals into N subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only exist when the number of segments are fixed.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights in
            case of Gaussian quadrature scheme. If omitted, ``analytical`` will
            be tried first, and ``stieltjes`` used if that fails.
        tolerance (float):
            The allowed relative error in norm between two quadrature orders
            before method assumes convergence.
        scaling (float):
            A multiplier the adaptive order increases with for each step
            quadrature order is not converged. Use 0 to indicate unit
            increments.
        n_max (int):
            The allowed number of quadrature points to use in approximation.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Abscissas and weights created from full tensor grid rule. Flatten
            such that ``abscissas.shape == (len(dist), len(weights))``.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)
        >>> abscissas, weights = generate_quadrature(
        ...     1, distribution, rule=("gaussian", "fejer_2"))
        >>> abscissas.round(3)
        array([[-1.  , -1.  ,  1.  ,  1.  ],
               [-4.11,  4.11, -4.11,  4.11]])
        >>> weights.round(3)
        array([0.222, 0.222, 0.222, 0.222])

    See also:
        :func:`chaospy.quadrature.clenshaw_curtis`
        :func:`chaospy.quadrature.fejer_1`
        :func:`chaospy.quadrature.fejer_2`
        :func:`chaospy.quadrature.gaussian`
        :func:`chaospy.quadrature.legendre_proxy`
        :func:`chaospy.quadrature.lobatto`
        :func:`chaospy.quadrature.kronrod`
        :func:`chaospy.quadrature.patterson`
        :func:`chaospy.quadrature.radau`
        :func:`chaospy.quadrature.leja`
        :func:`chaospy.quadrature.newton_cotes`
        :func:`chaospy.quadrature.discrete`
        :func:`chaospy.quadrature.grid`
        :func:`chaospy.quadrature.genz_keister_16`
        :func:`chaospy.quadrature.genz_keister_18`
        :func:`chaospy.quadrature.genz_keister_22`
        :func:`chaospy.quadrature.genz_keister_24`

    """
    if not rule:
        rule = [
            ("discrete" if dist_.interpret_as_integer else "clenshaw_curtis")
            for dist_ in dist
        ]

    if sparse:
        return sparse_grid(
            order=order,
            dist=dist,
            growth=growth,
            recurrence_algorithm=recurrence_algorithm,
            rule=rule,
            tolerance=tolerance,
            scaling=scaling,
            n_max=n_max,
        )

    if (not isinstance(dist, chaospy.Distribution) or
            (len(dist) == 1 or dist.stochastic_dependent)):
        if not isinstance(rule, str) and len(set(rule)) == 1:
            rule = rule[0]
        assert isinstance(rule, str), (
            "dependencies require rule consistency; %s provided" % rule)
        abscissas, weights = _generate_quadrature(
            order=order,
            dist=dist,
            rule=rule,
            growth=growth,
            segments=segments,
            recurrence_algorithm=recurrence_algorithm,
            tolerance=tolerance,
            scaling=scaling,
            n_max=n_max,
        )

    else:
        if isinstance(rule, str):
            rule = [rule]*len(dist)
        assert len(rule) == len(dist), (
            "rules and distribution length does not match.")
        assert all(isinstance(rule_, str) for rule_ in rule)

        order = numpy.ones(len(dist), dtype=int)*order
        abscissas, weights = zip(*[
            _generate_quadrature(
                order=order_,
                dist=dist_,
                rule=rule_,
                growth=growth,
                segments=segments,
                recurrence_algorithm=recurrence_algorithm,
                tolerance=tolerance,
                scaling=scaling,
                n_max=n_max,
            )
            for order_, dist_, rule_ in zip(order, dist, rule)
        ])
        abscissas = combine([abscissa.T for abscissa in abscissas]).T
        weights = numpy.prod(combine([weight.T for weight in weights]), -1)

    if isinstance(dist, chaospy.Distribution):
        assert abscissas.shape == (len(dist), len(weights))
        if dist.interpret_as_integer:
            abscissas = numpy.round(abscissas).astype(int)
    else:
        assert abscissas.shape[-1] == len(weights)
    return abscissas, weights


def _generate_quadrature(order, dist, rule, **kwargs):

    logger = logging.getLogger(__name__)
    if isinstance(dist, chaospy.OperatorDistribution):

        args = ("left", "right")
        right_dist = isinstance(dist._parameters["right"], chaospy.Distribution)
        args = args if right_dist else args[::-1]
        assert not isinstance(dist._parameters[args[0]], chaospy.Distribution)
        const = numpy.asarray(dist._parameters[args[0]])

        if isinstance(dist, chaospy.Add):
            dist = dist._parameters[args[1]]
            abscissas, weights = _generate_quadrature(
                order=order, dist=dist, rule=rule, **kwargs)
            abscissas = (abscissas.T+const.T).T
            return abscissas, weights

        elif isinstance(dist, chaospy.Multiply):
            dist = dist._parameters[args[1]]
            abscissas, weights = _generate_quadrature(
                order=order, dist=dist, rule=rule, **kwargs)
            abscissas = (abscissas.T*const.T).T
            return abscissas, weights

    rule = rule.lower()
    if rule in DEPRECATED_SHORT_NAMES:
        logger.warning("quadrature rule '%s' is renamed to '%s'; "
                       "error will be raised in the future",
                       rule, DEPRECATED_SHORT_NAMES[rule])
        rule = DEPRECATED_SHORT_NAMES[rule]
    rule = SHORT_NAME_TABLE[rule]

    parameters = {}

    if rule in ("clenshaw_curtis", "fejer_1", "fejer_2", "newton_cotes", "discrete", "grid"):
        parameters["growth"] = kwargs["growth"]

    if rule in ("clenshaw_curtis", "fejer_1", "fejer_2", "newton_cotes", "grid", "legendre"):
        parameters["segments"] = kwargs["segments"]

    if rule in ("gaussian", "kronrod", "radau", "lobatto"):
        parameters.update(
            n_max=kwargs["n_max"],
            tolerance=kwargs["tolerance"],
            scaling=kwargs["scaling"],
            recurrence_algorithm=kwargs["recurrence_algorithm"],
        )

    quad_function = chaospy.quadrature.INTEGRATION_COLLECTION[rule]
    abscissas, weights = quad_function(order, dist, **parameters)
    return abscissas, weights



"""
Truncation.

Example usage
-------------

Simple distribution to start with::

    >>> distribution = chaospy.Normal(0, 1)
    >>> distribution.inv([0.9, 0.99, 0.999]).round(4)
    array([1.2816, 2.3263, 3.0902])

Same distribution, but with a upper truncation::

    >>> right_trunc = chaospy.Trunc(chaospy.Normal(0, 1), upper=1)
    >>> right_trunc
    Trunc(Normal(mu=0, sigma=1), upper=1)
    >>> right_trunc.inv([0.9, 0.99, 0.999]).round(4)
    array([0.6974, 0.9658, 0.9965])

Same, but with lower truncation::

    >>> left_trunc = chaospy.Trunc(chaospy.Normal(0, 1), lower=1)
    >>> left_trunc
    Trunc(Normal(mu=0, sigma=1), lower=1)
    >>> left_trunc.inv([0.001, 0.01, 0.1]).round(4)
    array([1.0007, 1.0066, 1.0679])

"""
import numpy
import chaospy

from ..baseclass import Distribution, OperatorDistribution


class Trunc(Distribution):
    """Truncation."""

    def __init__(self, dist, lower=None, upper=None):
        """
        Constructor.

        Args:
            dist (Distribution):
                Distribution to be truncated.
            lower (Distribution, numpy.ndarray):
                Lower truncation bound.
            upper (Distribution, numpy.ndarray):
                Upper truncation bound.
        """
        assert isinstance(dist, Distribution)
        repr_args = [dist]
        repr_args += chaospy.format_repr_kwargs(lower=(lower, None))
        repr_args += chaospy.format_repr_kwargs(upper=(upper, None))
        exclusion = set()
        for deps in dist._dependencies:
            exclusion.update(deps)
        if isinstance(lower, Distribution):
            if lower.stochastic_dependent:
                raise chaospy.StochasticallyDependentError(
                    "Joint distribution with dependencies not supported.")
            assert len(dist) == len(lower)
            lower_ = lower.lower
        elif lower is None:
            lower = lower_ = dist.lower
        else:
            lower = lower_ = numpy.atleast_1d(lower)
        if isinstance(upper, Distribution):
            if upper.stochastic_dependent:
                raise chaospy.StochasticallyDependentError(
                    "Joint distribution with dependencies not supported.")
            assert len(dist) == len(upper)
            upper_ = upper.upper
        elif upper is None:
            upper = upper_ = dist.upper
        else:
            upper = upper_ = numpy.atleast_1d(upper)
        assert numpy.all(upper_ > lower_), (
            "condition `upper > lower` not satisfied: %s <= %s" % (upper_, lower_))

        dependencies, parameters, rotation = chaospy.declare_dependencies(
            distribution=self,
            parameters=dict(lower=lower, upper=upper),
            length=len(dist),
        )
        super(Trunc, self).__init__(
            parameters=parameters,
            dependencies=dependencies,
            exclusion=exclusion,
            repr_args=repr_args,
        )
        self._dist = dist

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(Trunc, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        assert set(parameters) == {"cache", "lower", "upper", "idx"}

        if isinstance(parameters["lower"], Distribution):
            parameters["lower"] = parameters["lower"]._get_cache(idx, cache=parameters["cache"], get=0)
        elif len(parameters["lower"]) > 1 and idx is not None:
            parameters["lower"] = parameters["lower"][idx]
        if isinstance(parameters["upper"], Distribution):
            parameters["upper"] = parameters["upper"]._get_cache(idx, cache=parameters["cache"], get=0)
        elif len(parameters["upper"]) > 1 and idx is not None:
            parameters["upper"] = parameters["upper"][idx]
        if assert_numerical:
            assert (not isinstance(parameters["lower"], Distribution) or
                    not isinstance(parameters["upper"], Distribution))
        if idx is None:
            del parameters["idx"]
        return parameters

    def _lower(self, idx, lower, upper, cache):
        """
        Distribution lower bound.

        Examples:
            >>> chaospy.Trunc(chaospy.Uniform(), upper=0.6).lower
            array([0.])
            >>> chaospy.Trunc(chaospy.Uniform(), lower=0.6).lower
            array([0.6])
        """
        del upper
        if isinstance(lower, Distribution):
            lower = lower._get_lower(idx, cache=cache)
        return lower

    def _upper(self, idx, lower, upper, cache):
        """
        Distribution lower bound.

        Examples:
            >>> chaospy.Trunc(chaospy.Uniform(), upper=0.6).upper
            array([0.6])
            >>> chaospy.Trunc(chaospy.Uniform(), lower=0.6).upper
            array([1.])
        """
        del lower
        if isinstance(upper, Distribution):
            upper = upper._get_upper(idx, cache=cache)
        return upper

    def _cdf(self, xloc, idx, lower, upper, cache):
        """
        Cumulative distribution function.

        Example:
            >>> chaospy.Uniform().fwd([-0.5, 0.3, 0.7, 1.2])
            array([0. , 0.3, 0.7, 1. ])
            >>> chaospy.Trunc(chaospy.Uniform(), upper=0.4).fwd([-0.5, 0.2, 0.8, 1.2])
            array([0. , 0.5, 1. , 1. ])
            >>> chaospy.Trunc(chaospy.Uniform(), lower=0.6).fwd([-0.5, 0.2, 0.8, 1.2])
            array([0. , 0. , 0.5, 1. ])
        """
        assert not isinstance(lower, Distribution)
        assert not isinstance(upper, Distribution)
        lower = numpy.broadcast_to(lower, xloc.shape)
        upper = numpy.broadcast_to(upper, xloc.shape)
        lower = self._dist._get_fwd(lower, idx, cache=cache.copy())
        upper = self._dist._get_fwd(upper, idx, cache=cache.copy())
        uloc = self._dist._get_fwd(xloc, idx, cache)
        return (uloc-lower)/(1-lower)/upper

    def _pdf(self, xloc, idx, lower, upper, cache):
        """
        Probability density function.

        Example:
            >>> dist = chaospy.Trunc(chaospy.Uniform(), upper=0.6)
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0.        , 1.66666667, 1.66666667, 0.        , 0.        ])
            >>> dist = chaospy.Trunc(chaospy.Uniform(), upper=0.4)
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0. , 2.5, 0. , 0. , 0. ])
            >>> dist = chaospy.Trunc(chaospy.Uniform(), lower=0.4)
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0.        , 0.        , 1.66666667, 1.66666667, 0.        ])
            >>> dist = chaospy.Trunc(chaospy.Uniform(), lower=0.6)
            >>> dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25])
            array([0. , 0. , 0. , 2.5, 0. ])
        """
        assert not isinstance(lower, Distribution)
        assert not isinstance(upper, Distribution)
        lower = numpy.broadcast_to(lower, xloc.shape)
        upper = numpy.broadcast_to(upper, xloc.shape)
        lower = self._dist._get_fwd(lower, idx, cache=cache.copy())
        upper = self._dist._get_fwd(upper, idx, cache=cache.copy())
        uloc = self._dist._get_pdf(xloc, idx, cache=cache)
        return uloc/(1-lower)/upper

    def _ppf(self, qloc, idx, lower, upper, cache):
        """
        Point percentile function.

        Example:
            >>> chaospy.Uniform().inv([0.1, 0.2, 0.9])
            array([0.1, 0.2, 0.9])
            >>> chaospy.Trunc(chaospy.Uniform(), upper=0.4).inv([0.1, 0.2, 0.9])
            array([0.04, 0.08, 0.36])
            >>> chaospy.Trunc(chaospy.Uniform(), lower=0.6).inv([0.1, 0.2, 0.9])
            array([0.64, 0.68, 0.96])
        """
        assert not isinstance(lower, Distribution)
        assert not isinstance(upper, Distribution)
        lower = numpy.broadcast_to(lower, qloc.shape)
        upper = numpy.broadcast_to(upper, qloc.shape)
        lower = self._dist._get_fwd(lower, idx, cache=cache.copy())
        upper = self._dist._get_fwd(upper, idx, cache=cache.copy())
        return self._dist._get_inv(qloc*upper*(1-lower)+lower, idx, cache=cache)



"""Generate the quadrature nodes and weights in Clenshaw-Curtis quadrature."""
try:
    from functools import lru_cache
except ImportError:  # pragma: no coverage
    from functools32 import lru_cache

import numpy
import chaospy

from .hypercube import hypercube_quadrature


def clenshaw_curtis(order, domain=(0., 1.), growth=False, segments=1):
    """
    Generate the quadrature nodes and weights in Clenshaw-Curtis quadrature.

    Clenshaw-Curtis quadrature method is a good all-around quadrature method
    comparable to Gaussian quadrature, but typically limited to finite
    intervals without a specific weight function. In addition to be quite
    accurate, the weights and abscissas can be calculated quite fast.

    Another thing to note is that Clenshaw-Curtis, with an appropriate growth
    rule is fully nested. This means, if one applies a method that combines
    different order of quadrature rules, the number of evaluations can often be
    reduced as the abscissas can be used across levels.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (:class:`chaospy.Distribution`, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.
        segments (int):
            Split intervals into steps subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only appear when the number of segments are fixed.

    Returns:
        abscissas (numpy.ndarray):
            The quadrature points for where to evaluate the model function
            with ``abscissas.shape == (len(dist), steps)`` where ``steps`` is
            the number of samples.
        weights (numpy.ndarray):
            The quadrature weights with ``weights.shape == (steps,)``.

    Notes:
        Implemented as proposed by Waldvogel :cite:`waldvogel_fast_2006`.

    Example:
        >>> abscissas, weights = chaospy.quadrature.clenshaw_curtis(4, (0, 1))
        >>> abscissas.round(4)
        array([[0.    , 0.1464, 0.5   , 0.8536, 1.    ]])
        >>> weights.round(4)
        array([0.0333, 0.2667, 0.4   , 0.2667, 0.0333])

    See also:
        :func:`chaospy.quadrature.gaussian`
        :func:`chaospy.quadrature.fejer_1`
        :func:`chaospy.quadrature.fejer_2`

    """
    order = numpy.asarray(order)
    order = numpy.where(growth, numpy.where(order > 0, 2**order, 0), order)
    return hypercube_quadrature(
        quad_func=clenshaw_curtis_simple,
        order=order,
        domain=domain,
        segments=segments,
    )


@lru_cache(None)
def clenshaw_curtis_simple(order):
    """
    Backend for Clenshaw-Curtis quadrature.

    Use :func:`chaospy.quadrature.clenshaw_curtis` instead.
    """
    order = int(order)
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])
    elif order == 1:
        return numpy.array([0., 1.]), numpy.array([0.5, 0.5])

    theta = (order-numpy.arange(order+1))*numpy.pi/order
    abscissas = 0.5*numpy.cos(theta)+0.5

    steps = numpy.arange(1, order, 2)
    length = len(steps)
    remains = order-length

    beta = numpy.hstack([2./(steps*(steps-2)), [1./steps[-1]], numpy.zeros(remains)])
    beta = -beta[:-1]-beta[:0:-1]

    gamma = -numpy.ones(order)
    gamma[length] += order
    gamma[remains] += order
    gamma /= (order**2-1+(order%2))

    weights = numpy.fft.ihfft(beta+gamma)
    assert max(weights.imag) < 1e-15
    weights = weights.real
    weights = numpy.hstack([weights, weights[len(weights)-2+(order%2)::-1]])/2
    assert numpy.isclose(numpy.sum(weights), 1)

    return abscissas, weights



"""
In some cases a constructed distribution that are first and foremost data
driven. In such scenarios it make sense to make use of
`kernel density estimation`_ (KDE). In ``chaospy`` KDE can be accessed through
the :func:`GaussianKDE` constructor.

Basic usage of the :func:`GaussianKDE` constructor involves just passing the
data as input argument::

    >>> data = [3, 4, 5, 5]
    >>> distribution = chaospy.GaussianKDE(data)

This distribution can be used as any other distributions::

    >>> distribution.cdf([3, 3.5, 4, 4.5, 5]).round(4)
    array([0.1393, 0.2542, 0.3889, 0.5512, 0.7359])
    >>> distribution.mom(1).round(4)
    4.25
    >>> distribution.sample(4).round(4)
    array([4.7784, 2.8769, 5.8109, 4.2995])

In addition multivariate distributions supported::

    >>> data = [[1, 2, 2, 3], [5, 5, 4, 3]]
    >>> distribution = chaospy.GaussianKDE(data)
    >>> distribution.sample(4).round(4)
    array([[2.081 , 3.0304, 3.0882, 0.4872],
           [3.2878, 2.5473, 2.2699, 5.3412]])

.. _kernel density estimation: \
https://en.wikipedia.org/wiki/Kernel_density_estimation
"""
from .gaussian import GaussianKDE
from .mixture import GaussianMixture



"""Distribution with user-provided methods."""
import numpy
import chaospy

from .simple import SimpleDistribution


class UserDistribution(SimpleDistribution):
    """
    Distribution with user-provided methods.

    The internals of this distribution is provided in the constructor.

    Examples:
        >>> def cdf(x_loc, lo, up):
        ...     '''Cumulative distribution function.'''
        ...     return (x_loc-lo)/(up-lo)
        >>> def pdf(x_loc, lo, up):
        ...     '''Probability density function.'''
        ...     return 1./(up-lo)
        >>> def lower(lo, up):
        ...     '''Lower bounds function.'''
        ...     return lo
        >>> def upper(lo, up):
        ...     '''Upper bounds function.'''
        ...     return up
        >>> distribution = chaospy.UserDistribution(
        ...     cdf, pdf, lower, upper, parameters=dict(lo=-1, up=1))
        >>> distribution
        UserDistribution(<function ..., parameters=dict(lo=-1, up=1))
        >>> distribution.fwd(numpy.linspace(-2, 2, 7)).round(4)
        array([0.    , 0.    , 0.1667, 0.5   , 0.8333, 1.    , 1.    ])
        >>> distribution.pdf(numpy.linspace(-2, 2, 7)).round(4)
        array([0. , 0. , 0.5, 0.5, 0.5, 0. , 0. ])
        >>> distribution.inv(numpy.linspace(0, 1, 7)).round(4)
        array([-1.    , -0.6667, -0.3333,  0.    ,  0.3333,  0.6667,  1.    ])
        >>> distribution.lower, distribution.upper
        (array([-1.]), array([1.]))

    """

    _cdf = None  # required by class constructor

    def __init__(
        self,
        cdf,
        pdf=None,
        lower=None,
        upper=None,
        ppf=None,
        mom=None,
        ttr=None,
        parameters=None
    ):
        """
        Args:
            cdf (Callable[[numpy.ndarray, ...], numpy.ndarray]):
                Cumulative distribution function.
            pdf (Callable[[numpy.ndarray, ...], numpy.ndarray]):
                Probability density function.
            lower (Callable[[...], numpy.ndarray]):
                Lower boundary.
            upper (Callable[[...], numpy.ndarray]):
                Upper boundary.
            ppf (Callable[[numpy.ndarray, ...], numpy.ndarray]):
                Point percentile function.
            mom (Callable[[numpy.ndarray, ...], numpy.ndarray]):
                Raw moment generator.
            ttr (Callable[[numpy.ndarray, ...], numpy.ndarray]):
                Three terms recurrence coefficient generator.
            parameters (Dict[str, numpy.ndarray]):
                Parameters to pass to each of the distribution methods.

        """
        self._cdf = cdf
        repr_args = [str(cdf)]
        if ppf is None and (lower is None or upper is None):
            raise chaospy.UnsupportedFeature("either ppf or lower+upper should be provided.")
        if pdf is not None:
            repr_args.append("pdf=%s" % pdf)
            self._pdf = pdf
        if lower is not None:
            repr_args.append("lower=%s" % lower)
            self._lower = lower
        if upper is not None:
            repr_args.append("upper=%s" % upper)
            self._upper = upper
        if ppf is not None:
            repr_args.append("ppf=%s" % ppf)
            self._ppf = ppf
        if mom is not None:
            repr_args.append("mom=%s" % mom)
            self._mom = mom
        if ttr is not None:
            repr_args.append("ttr=%s" % ttr)
            self._ttr = ttr

        parameters = parameters if parameters else {}
        if parameters:
            params = [("%s=%s" % (key, parameters[key]))
                        for key in sorted(parameters)]
            repr_args.append("parameters=dict(%s)" % ", ".join(params))

        super(UserDistribution, self).__init__(
            repr_args=repr_args, parameters=parameters)

    def _lower(self, **parameters):
        x_loc = 0.
        for param in parameters.values():
            x_loc, _ = numpy.broadcast_arrays(x_loc, param)
        return self._ppf(x_loc, **parameters)

    def _upper(self, **parameters):
        x_loc = 1.
        for param in parameters.values():
            x_loc, _ = numpy.broadcast_arrays(x_loc, param)
        return self._ppf(x_loc, **parameters)



r"""Create Gaussian quadrature nodes and weights."""
import chaospy

from .utils import combine_quadrature


def gaussian(
        order,
        dist,
        recurrence_algorithm="stieltjes",
        rule="clenshaw_curtis",
        tolerance=1e-10,
        scaling=3,
        n_max=5000,
):
    """
    Create Gaussian quadrature nodes and weights.

    Generating Gaussian quadrature by first generating so called *three terms
    recurrence* coefficients using one various different algorithms. The
    coefficients are them converted to abscissas and weights by constructing
    lower banded Jacobi matrix and calculating eigenvalues and eigenvectors,
    which can be directly translated to abscissas and weights. Construction of
    the coefficients is potentially numerically unstable, there for multiple
    algorithms exists.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.Distribution):
            The distribution which density will be used as weight function.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        tolerance (float):
            The allowed relative error in norm between two quadrature orders
            before method assumes convergence.
        scaling (float):
            A multiplier the adaptive order increases with for each step
            quadrature order is not converged. Use 0 to indicate unit
            increments.
        n_max (int):
            The allowed number of quadrature points to use in approximation.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Gaussian quadrature abscissas and weights with shapes
            ``(len(dist), order+1)`` and ``(order+1,)`` respectively.

    Raises:
        NotImplementedError:
            In the case of ``analytical`` three terms recurrence algorithm,
            error is raised if the distribution does not implement the feature.
            coefficients.
        numpy.linalg.LinAlgError:
            For non-canonical random variables, the construction might fail
            because of illegal numerical operations.

    Notes:
        Quadrature constructed as outlined by Walter Gautschi
        :cite:`gautschi_construction_1968`.

    Examples:
        >>> distribution = chaospy.Normal(0, 1)
        >>> abscissas, weights = chaospy.quadrature.gaussian(5, distribution)
        >>> abscissas.round(4)
        array([[-3.3243, -1.8892, -0.6167,  0.6167,  1.8892,  3.3243]])
        >>> weights.round(4)
        array([0.0026, 0.0886, 0.4088, 0.4088, 0.0886, 0.0026])
        >>> distribution = chaospy.J(chaospy.Uniform(), chaospy.Normal())
        >>> abscissas, weights = chaospy.quadrature.gaussian(2, distribution)
        >>> abscissas.round(2)
        array([[ 0.11,  0.11,  0.11,  0.5 ,  0.5 ,  0.5 ,  0.89,  0.89,  0.89],
               [-1.73,  0.  ,  1.73, -1.73,  0.  ,  1.73, -1.73,  0.  ,  1.73]])
        >>> weights.round(3)
        array([0.046, 0.185, 0.046, 0.074, 0.296, 0.074, 0.046, 0.185, 0.046])

    """
    coefficients = chaospy.construct_recurrence_coefficients(
        order=order,
        dist=dist,
        recurrence_algorithm=recurrence_algorithm,
        rule=rule,
        tolerance=tolerance,
        scaling=scaling,
        n_max=n_max,
    )
    abscissas, weights = chaospy.coefficients_to_quadrature(coefficients)
    return combine_quadrature(abscissas, weights)



"""
Laja quadrature is a newer method for performing quadrature in stochastical
problems. The method is described in a `journal paper`_ by Narayan and Jakeman.

.. _journal paper: https://arxiv.org/pdf/1404.5663.pdf

Example usage
-------------

The first few orders::

    >>> distribution = chaospy.Beta(2, 3)
    >>> for order in [0, 1, 2, 3, 4]:  # doctest: +NORMALIZE_WHITESPACE
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="leja")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.4]] [1.]
    1 [[0.11 0.4 ]] [0. 1.]
    2 [[0.11  0.4   0.787]] [0.203 0.644 0.153]
    3 [[0.11  0.4   0.6   0.787]] [0.22  0.574 0.086 0.12 ]
    4 [[0.027 0.11  0.4   0.6   0.787]] [-0.047  0.298  0.499  0.146  0.103]

"""
import numpy
from scipy.optimize import fminbound
import chaospy

from .utils import combine


def leja(
        order,
        dist,
        rule="fejer_2",
):
    """
    Generate Leja quadrature node.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution which density will be used as weight function.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Notes:
        Implemented as proposed in Narayan and Jakeman
        :cite:`narayan_adaptive_2014`.

    Example:
        >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)
        >>> abscissas, weights = chaospy.quadrature.leja(2, distribution)
        >>> abscissas.round(2)
        array([[-1.41, -1.41, -1.41,  0.  ,  0.  ,  0.  ,  1.76,  1.76,  1.76],
               [-1.41,  0.  ,  1.76, -1.41,  0.  ,  1.76, -1.41,  0.  ,  1.76]])
        >>> weights.round(3)
        array([0.05 , 0.133, 0.04 , 0.133, 0.359, 0.107, 0.04 , 0.107, 0.032])

    """
    if len(dist) > 1:
        if dist.stochastic_dependent:
            raise chaospy.StochasticallyDependentError(
                "Leja quadrature do not supper distribution with dependencies.")
        order = numpy.broadcast_to(order, len(dist))
        out = [leja(order[_], dist[_]) for _ in range(len(dist))]
        abscissas = [_[0][0] for _ in out]
        weights = [_[1] for _ in out]
        abscissas = combine(abscissas).T
        weights = combine(weights)
        weights = numpy.prod(weights, -1)

        return abscissas, weights

    abscissas = [dist.lower, dist.mom(1).flatten(), dist.upper]
    for _ in range(int(order)):

        def objective(abscissas_):
            """Local objective function."""
            out = -numpy.sqrt(dist.pdf(abscissas_))*numpy.prod(
                numpy.abs(abscissas[1:-1]-abscissas_))
            return out

        def fmin(idx):
            """Bound minimization."""
            try:
                xopt, fval, _, _ = fminbound(objective, abscissas[idx],
                                             abscissas[idx+1], full_output=True)
            # Hard coded solution to scipy/scipy#11207 for scipy < 1.5.0.
            except UnboundLocalError:  # pragma: no cover
                xopt = abscissas[idx]+0.5*(3-5**0.5)*(
                    abscissas[idx+1]-abscissas[idx])
                fx = objective(xopt)
            return xopt, fval

        opts, vals = zip(*[fmin(idx) for idx in range(len(abscissas)-1)])
        index = numpy.argmin(vals)
        abscissas.insert(index+1, opts[index])

    abscissas = numpy.asfarray(abscissas).flatten()[1:-1]
    weights = create_weights(abscissas, dist, rule)
    abscissas = abscissas.reshape(1, abscissas.size)

    return numpy.asfarray(abscissas), numpy.asfarray(weights).flatten()


def create_weights(
        nodes,
        dist,
        rule="clenshaw_curtis",
):
    """Create weights for the Laja method."""
    _, poly, _ = chaospy.stieltjes(len(nodes)-1, dist, rule=rule)
    poly = poly.ravel()
    weights = numpy.linalg.inv(poly(nodes))
    return weights[:, 0]



"""Construct recurrence coefficients."""
import numpy
import chaospy

from .chebyshev import modified_chebyshev
from .jacobi import coefficients_to_quadrature
from .lanczos import lanczos
from .stieltjes import stieltjes

RECURRENCE_ALGORITHMS = ("chebyshev", "lanczos", "stieltjes")


def construct_recurrence_coefficients(
        order,
        dist,
        recurrence_algorithm="stieltjes",
        rule="clenshaw_curtis",
        tolerance=1e-10,
        scaling=3,
        n_max=5000,
):
    """
    Frontend wrapper for constructing *three terms recurrence* coefficients.

    The algorithm for constructing recurrence coefficients can be specified
    using the ``recurrence_algorithm`` flag. It accepts the strings:

    ``stieltjes``
        Use the discretized Stieltjes algorithm for iterative estimate each
        recurrence coefficient using numerical integration. Typically the
        method known to have the highest convergence rate.
    ``chebyshev``
        Use modified Chebyshev algorithm to convert raw statistical moments to
        the recurrence coefficients. A good algorithm for when raw statistical
        moments are known analytically.
    ``lanczos``
        Use a known relationship between the Jakobi matrix and a matrix
        consisting of abscissas and weights from an alternative integration
        scheme to estimate the recurrence coefficients. Stabilized using ideas
        by Rutishauser. An alternative method to ``stieltjes``.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution which density will be used as weight function.
            Assumed to one-dimensional.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        tolerance (float):
            The allowed relative error in norm between two quadrature orders
            before method assumes convergence.
        scaling (float):
            A multiplier the adaptive order increases with for each step
            quadrature order is not converged. Use 0 to indicate unit
            increments.
        n_max (int):
            The allowed number of quadrature points to use in approximation.

    Returns:
        (typing.List[numpy.ndarray]):
            List of recurrence coefficients with shape ``(2, order+1)``. The
            alpha and beta coefficients can found in ``out[0]`` and ``out[1]``
            respectively.

    Examples:
        >>> distribution = chaospy.Normal(1, 1)
        >>> coefficients = chaospy.construct_recurrence_coefficients(
        ...     4, distribution, recurrence_algorithm="stieltjes")
        >>> coefficients[0].round(3)
        array([[1., 1., 1., 1., 1.],
               [1., 1., 2., 3., 4.]])
        >>> distribution = chaospy.J(chaospy.Exponential(), chaospy.Uniform())
        >>> coefficients = chaospy.construct_recurrence_coefficients(
        ...     [2, 4], distribution, recurrence_algorithm="chebyshev")
        >>> coefficients[0].round(4)
        array([[1., 3., 5.],
               [1., 1., 4.]])
        >>> coefficients[1].round(4)
        array([[0.5   , 0.5   , 0.5   , 0.5   , 0.5   ],
               [1.    , 0.0833, 0.0667, 0.0643, 0.0635]])
    """
    assert isinstance(dist, chaospy.Distribution), (
        "%s is not a distribution" % str(dist))
    if len(dist) > 1:
        orders = (order*numpy.ones(len(dist), dtype=int)).tolist()
        return [construct_recurrence_coefficients(
            order=int(order_),
            dist=dist_,
            recurrence_algorithm=recurrence_algorithm,
            rule=rule,
            tolerance=tolerance,
            scaling=scaling,
            n_max=n_max,
        )[0] for dist_, order_ in zip(dist, orders)]

    assert recurrence_algorithm in RECURRENCE_ALGORITHMS, (
        "recurrence algorithm '%s' not recognized" % recurrence_algorithm)
    assert not rule.startswith("gauss"), (
        "recursive Gaussian quadrature construct")

    if recurrence_algorithm == "chebyshev":
        moments = dist.mom(numpy.arange(2*(order+1), dtype=int))
        coeffs = modified_chebyshev(moments)

    elif recurrence_algorithm == "lanczos":
        coeffs = lanczos(order, dist, rule=rule, tolerance=tolerance)

    elif recurrence_algorithm == "stieltjes":
        coeffs, _, _ = stieltjes(order, dist, rule=rule, tolerance=tolerance)

    return [coeffs.reshape(2, int(order)+1)]



import numpy
import chaospy


def legendre(
        order,
        lower=-1,
        upper=1,
        physicist=False,
        normed=False,
        retall=False,
):
    """
    Examples:
        >>> polynomials, norms = chaospy.expansion.legendre(3, retall=True)
        >>> polynomials.round(5)
        polynomial([1.0, q0, q0**2-0.33333, q0**3-0.6*q0])
        >>> norms
        array([1.        , 0.33333333, 0.08888889, 0.02285714])
        >>> chaospy.expansion.legendre(3, physicist=True).round(3)
        polynomial([1.0, 1.5*q0, 2.5*q0**2-0.556, 4.375*q0**3-1.672*q0])
        >>> chaospy.expansion.legendre(3, lower=0, upper=1, normed=True).round(3)
        polynomial([1.0, 3.464*q0-1.732, 13.416*q0**2-13.416*q0+2.236,
                    52.915*q0**3-79.373*q0**2+31.749*q0-2.646])

    """
    multiplier = 1.
    if physicist:
        multiplier = numpy.arange(1, order+1)
        multiplier = (2*multiplier+1)/(multiplier+1)
    _, [polynomials], [norms] = chaospy.recurrence.analytical_stieltjes(
        order, chaospy.Uniform(lower, upper), multiplier=multiplier)
    if normed:
        polynomials = chaospy.true_divide(polynomials, numpy.sqrt(norms))
        norms[:] = 1.
    return (polynomials, norms) if retall else polynomials
